import functools
import numpy as np
import warnings

import matplotlib.axes
from pyproj import Geod

from .skycrs import PlateCarreeCRS
from .utils import wrap_values

__all__ = ["SkyAxes"]


def _add_lonlat(func):
    """Decorator to add lonlat option."""
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if kwargs.pop('lonlat', True):
            kwargs['transform'] = self.projection
        return func(self, *args, **kwargs)
    return wrapper


class SkyAxes(matplotlib.axes.Axes):
    # docstring inherited
    def __init__(self, *args, **kwargs):
        self.projection = kwargs.pop("sky_crs")

        self.plate_carree = PlateCarreeCRS()

        super().__init__(*args, **kwargs)

    def clear(self):
        """Clear the current axes."""
        result = super().clear()
        self.xaxis.set_visible(False)
        self.yaxis.set_visible(False)

        self.set_frame_on(False)

        # Always equal aspect ratio.
        self.set_aspect('equal')

        return result

    def cla_mplpre36(self):
        """Clear the current axes (deprecated in mpl 3.6)."""
        result = super().cla()
        self.xaxis.set_visible(False)
        self.yaxis.set_visible(False)

        self.set_frame_on(False)

        # Always equal aspect ratio.
        self.set_aspect('equal')

        return result

    def set_extent(self, extent, lonlat=True):
        """Set the extent of the axes.

        Parameters
        ----------
        extent : `tuple` [`float`]
            Set the extent by [lon0, lon1, lat0, lat1] or [x0, x1, y0, y1].
        lonlat : `bool`, optional
            Extent is specified in lon/lat coordinates?  Otherwise native.
        """
        if not lonlat:
            x0, x1, y0, y1 = extent
        else:
            # Need to do transformation.
            lon0, lon1, lat0, lat1 = extent

            # Check if the longitude range is the full sphere, and special case that.
            if np.isclose(np.abs(lon1 - lon0), 360.0):
                lat_steps = [lat0, lat1]
                if lat0 < 0.0 and lat1 > 0.0:
                    # Make sure we have the equator included
                    lat_steps.append(0.0)
                lon, lat = np.meshgrid(np.linspace(0, 360.0, 360), lat_steps)
                xy = self.projection.transform_points(self.plate_carree, lon.ravel(), lat.ravel())
                # Need to offset this by some small amount to ensure we don't get
                # out-of-bounds transformations.
                eps = 1e-5
                x0 = (1 - eps)*np.min(xy[:, 0])
                x1 = (1 - eps)*np.max(xy[:, 0])
                y0 = (1 - eps)*np.min(xy[:, 1])
                y1 = (1 - eps)*np.max(xy[:, 1])
            else:
                # Make a ring of points and check their extent.
                npt = 100
                lon_pts = np.linspace(lon0, lon1, npt)
                lat_pts = np.linspace(lat0, lat1, npt)
                lon = np.concatenate((lon_pts, lon_pts, np.repeat(lon0, npt), np.repeat(lon1, npt)))
                lat = np.concatenate((np.repeat(lat0, npt), np.repeat(lat1, npt), lat_pts, lat_pts))
                xy = self.projection.transform_points(self.plate_carree, lon, lat)
                # FIXME NOTE NEED TO KNOW LON_0/WRAP OF PROJECTION...
                x0 = np.min(xy[:, 0])
                x1 = np.max(xy[:, 0])
                y0 = np.min(xy[:, 1])
                y1 = np.max(xy[:, 1])

        self.set_xlim([x0, x1])
        self.set_ylim([y0, y1])

    def get_extent(self, lonlat=True):
        """Get the extent of the axes.

        Parameters
        ----------
        lonlat : `bool`, optional
            Return extent in lon/lat coordinates?  Otherwise native.
        """
        x0, x1 = self.get_xlim()
        y0, y1 = self.get_ylim()

        if not lonlat:
            extent = (x0, x1, y0, y1)
        else:
            # Make a ring of points and check their extent.
            npt = 500
            x_pts = np.linspace(x0, x1, npt)
            y_pts = np.linspace(y0, y1, npt)
            x = np.concatenate((x_pts, x_pts, np.repeat(x0, npt), np.repeat(x1, npt)))
            y = np.concatenate((np.repeat(y0, npt), np.repeat(y1, npt), y_pts, y_pts))
            lonlat = self.plate_carree.transform_points(self.projection, x, y)

            # Check for out-of-bounds by reverse-projecting
            xy = self.projection.transform_points(self.plate_carree, lonlat[:, 0], lonlat[:, 1])
            bad = ((~np.isclose(xy[:, 0], x)) | (~np.isclose(xy[:, 1], y)))
            lonlat[bad, :] = np.nan

            # We need to wrap values to get the correct range
            wrap = (self.projection.lon_0 + 180.) % 360.
            with warnings.catch_warnings():
                # Some longitude values may be nan, so we filter these expected warnings.
                warnings.simplefilter("ignore")
                lon_wrap = wrap_values(lonlat[:, 0], wrap)

            if np.all(np.isnan(lon_wrap)):
                lon0 = -180.0 + 1e-5
                lon1 = 180.0 - 1e-5
            else:
                lon0 = np.nanmin(lon_wrap)
                lon1 = np.nanmax(lon_wrap)

            if np.all(np.isnan(lonlat[:, 1])):
                lat0 = -90.0 + 1e-5
                lat1 = 90.0 - 1e-5
            else:
                lat0 = np.nanmin(lonlat[:, 1])
                lat1 = np.nanmax(lonlat[:, 1])

            if self.xaxis_inverted():
                extent = (lon1, lon0, lat0, lat1)
            else:
                extent = (lon0, lon1, lat0, lat1)

        return extent

    @_add_lonlat
    def plot(self, *args, **kwargs):
        # docstring inherited

        # The transformation code will automatically plot geodesics
        # and split line segements that cross the wrapping boundary.
        result = super().plot(*args, **kwargs)

        return result

    @_add_lonlat
    def scatter(self, *args, **kwargs):
        # docstring inherited
        result = super().scatter(*args, **kwargs)

        return result

    @_add_lonlat
    def pcolormesh(self, X, Y, C, **kwargs):
        # docstring inherited
        C_temp = C.copy()

        if kwargs.get('lonlat', True):
            # Check for wrapping by projecting and looking for jumps.
            proj_xy = self.projection.transform_points(self.plate_carree, X, Y)
            X_proj = proj_xy[..., 0]
            Y_proj = proj_xy[..., 1]

            dist = np.hypot(X_proj[1:, 1:] - X_proj[0: -1, 0: -1],
                            Y_proj[1:, 1:] - Y_proj[0: -1, 0: -1])

            # If we have a jump of 10% of the radius, assume it's bad,
            # except if we are using PlateCarree which doesn't use the radius.
            if self.projection == self.plate_carree:
                max_dist = 90.0
            else:
                max_dist = 0.1*self.projection.radius

            split = (dist > max_dist).nonzero()

            # By marking these jumps as masked then pcolormesh works just fine.
            C_temp.mask[split] = True

        result = super().pcolormesh(X, Y, C_temp, **kwargs)

        return result

    @_add_lonlat
    def fill(self, *args, **kwargs):
        # docstring inherited
        result = super().fill(*args, **kwargs)

        return result

    @_add_lonlat
    def text(self, *args, **kwargs):
        # docstring inherited
        result = super().text(*args, **kwargs)

        return result

    @_add_lonlat
    def circle(self, lon, lat, radius, nsamp=100, fill=False, **kwargs):
        """Draw a geodesic circle centered at given position.

        Parameters
        ----------
        lon : `float`
            Longitude of center of circle (degrees).
        lat : `float`
            Latitude of center of circle (degrees).
        radius : `float`
            Radius of circle (degrees).
        nsamp : `int`, optional
            Number of points to sample.
        fill : `bool`, optional
            Draw filled circle?
        **kwargs : `dict`
            Extra plotting kwargs.
        """
        geod = Geod(a=self.projection.radius)

        # We need the radius in meters
        radius_m = self.projection.radius*np.deg2rad(radius)

        az = np.linspace(360.0, 0.0, nsamp)
        lons, lats, _ = geod.fwd(
            np.full(nsamp, lon, dtype=np.float64),
            np.full(nsamp, lat, dtype=np.float64),
            az,
            np.full(nsamp, radius_m)
        )
        if fill:
            return self.fill(lons, lats, **kwargs)
        else:
            return self.plot(lons, lats, **kwargs)

    @_add_lonlat
    def ellipse(self, lon, lat, a, b, theta, nsamp=100, fill=False, **kwargs):
        """Draw a geodesic ellipse centered at given position.

        Parameters
        ----------
        lon : `float`
            Longitude of center of ellipse (degrees).
        lat : `float`
            Latitude of center of ellipse (degrees).
        a : `float`
            Semi-major axis of ellipse (degrees).
        b : `float`
            Semi-minor axis of ellipse (degrees).
        theta : `float`
            Position angle of ellipse.  Degrees East of North.
        nsamp : `int`, optional
            Number of points to sample.
        fill : `bool`, optional
            Draw filled ellipse?
        **kwargs : `dict`
            Extra plotting kwargs.
        """
        geod = Geod(a=self.projection.radius)

        # We need the radius in meters
        a_m = self.projection.radius * np.deg2rad(a)
        b_m = self.projection.radius * np.deg2rad(b)

        az = np.linspace(360.0, 0.0, nsamp)

        phase_rad = np.deg2rad(az - theta)

        # Position Angle is defined as degrees East from North
        denom = np.sqrt((b_m * np.cos(phase_rad))**2 + (a_m * np.sin(phase_rad))**2)
        dist = a_m * b_m / denom

        lons, lats, _ = geod.fwd(
            np.full(nsamp, lon, dtype=np.float64),
            np.full(nsamp, lat, dtype=np.float64),
            az,
            dist,
        )
        if fill:
            return self.fill(lons, lats, **kwargs)
        else:
            return self.plot(lons, lats, **kwargs)

    @property
    def lon_0(self):
        return self.projection.lon_0

    @property
    def lat_0(self):
        return self.projection.lat_0

    def update_projection(self, crs_new):
        """Update the projection central coordinate.

        Parameters
        ----------
        crs_new : `skyproj.SkyCRS`
        """
        self.projection = crs_new



import os
import pytest

import numpy as np
import healsparse as hsp

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


def _get_hspmap():
    # Make a square noise field
    np.random.seed(1234)

    hspmap = hsp.HealSparseMap.make_empty(32, 4096, np.float32)
    poly = hsp.geom.Polygon(ra=[0.0, 10.0, 10.0, 0.0], dec=[0.0, 0.0, 10.0, 10.0], value=1.0)
    pixels = poly.get_pixels(nside=hspmap.nside_sparse)
    hspmap[pixels] = np.random.normal(size=pixels.size).astype(np.float32)
    # Add in a central square of fixed value.
    poly2 = hsp.geom.Polygon(ra=[5, 5.2, 5.2, 5.0], dec=[5, 5.0, 5.2, 5.2], value=3.0)
    pixels2 = poly2.get_pixels(nside=hspmap.nside_sparse)
    hspmap[pixels2] = 3.0

    return hspmap


def _get_hspmap_bool():
    hspmap = hsp.HealSparseMap.make_empty(32, 4096, bool)
    # Create a region of True
    poly = hsp.geom.Polygon(ra=[0.0, 10.0, 10.0, 0.0], dec=[0.0, 0.0, 10.0, 10.0], value=1.0)
    pixels = poly.get_pixels(nside=hspmap.nside_sparse)
    hspmap[pixels] = np.full(shape=pixels.size, fill_value=True)
    # Create a central square of False within the True region.
    poly = hsp.geom.Polygon(ra=[5, 5.5, 5.5, 5.0], dec=[5, 5.0, 5.5, 5.5], value=3.0)
    pixels2 = poly.get_pixels(nside=hspmap.nside_sparse)
    hspmap[pixels2] = np.full(shape=pixels2.size, fill_value=False)

    return hspmap


def _get_hspmap_rec_array():
    # Make a square noise field
    np.random.seed(1234)

    dtype = [("A", np.int64), ("B", np.int64)]
    hspmap = hsp.HealSparseMap.make_empty(32, 4096, dtype=dtype, primary="A")
    poly = hsp.geom.Polygon(ra=[0.0, 10.0, 10.0, 0.0], dec=[0.0, 0.0, 10.0, 10.0], value=1.0)
    pixels = poly.get_pixels(nside=hspmap.nside_sparse)
    hspmap[pixels] = np.zeros(shape=pixels.size, dtype=dtype)
    hspmap["A"][pixels] = np.random.normal(size=pixels.size).astype(np.int64)
    hspmap["B"][pixels] = np.random.normal(size=pixels.size).astype(np.int64)
    # Add in a central square of fixed value.
    poly2 = hsp.geom.Polygon(ra=[5, 5.2, 5.2, 5.0], dec=[5, 5.0, 5.2, 5.2], value=3.0)
    pixels2 = poly2.get_pixels(nside=hspmap.nside_sparse)
    hspmap["A"][pixels2] = np.full(shape=pixels2.size, fill_value=3)
    hspmap["B"][pixels2] = np.full(shape=pixels2.size, fill_value=1)

    return hspmap


def test_healsparse(tmp_path):
    """Test plotting a healsparse map."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname = 'healsparse_one.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    # Try a different colormap on the zoom.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    sp.draw_hspmap(hspmap, lon_range=[4.9, 5.3], lat_range=[4.9, 5.3], cmap=plt.colormaps['rainbow'])
    sp.draw_colorbar()
    fname = 'healsparse_two.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax, lon_0=-175.0)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap, zoom=False)
    sp.draw_inset_colorbar()
    fname = 'healsparse_three.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap, valid_mask=True)
    sp.draw_inset_colorbar()
    fname = 'healsparse_valid_pixels.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.DESSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname = 'healsparse_four.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_healsparse_nanval(tmp_path):
    """Test plotting a healsparse map that has a nan value."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()

    hspmap[hspmap.valid_pixels[0]] = np.nan

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname = 'healsparse_one.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_healsparse_rasterize(tmp_path):
    """Test plotting a healsparse map with and without rasterization."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    # We use a small xsize here because otherwise the non-rasterized test pdf
    # (below) is 22Mb and the test takes a long time to run.
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap, rasterized=True, xsize=200, zoom=False)
    sp.draw_inset_colorbar()
    fname = 'healsparse_rasterized_on.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_on = os.path.getsize(tmp_path / fname)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(
        hspmap,
        rasterized=False,
        xsize=200,
        zoom=False,
    )
    sp.draw_inset_colorbar()
    fname = 'healsparse_rasterized_off.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_off = os.path.getsize(tmp_path / fname)

    # The rasterized map will be smaller than the non-rasterized map.
    assert size_rasterized_on < size_rasterized_off

    # Test with a redraw event.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap, rasterized=True, xsize=200, zoom=False)
    sp.set_extent([10.5, -0.5, -0.5, 10.5])
    sp.draw_inset_colorbar()
    fname = 'healsparse_rasterized_on2.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_on2 = os.path.getsize(tmp_path / fname)

    # The rasterized map will be smaller than the non-rasterized map.
    assert size_rasterized_on2 < size_rasterized_off


def test_healsparse_bool(tmp_path):
    """Test plotting a boolean healsparse map.
    """
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap_bool()

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname = 'healsparse_bool.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap, valid_mask=True)
    sp.draw_inset_colorbar()
    fname = 'healsparse_bool_valid_pixels.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_healsparse_empty(tmp_path):
    """Test plotting an empty map.
    """
    # This simply checks that the plot can be made to
    # avoid the empty autoscale bug.
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = hsp.HealSparseMap.make_empty(32, 4096, np.float32)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    _ = sp.draw_hspmap(
        hspmap,
        lon_range=[10, 20],
        lat_range=[10, 20],
        zoom=False,
    )


def test_healpix(tmp_path):
    """Test plotting a healpix map."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()
    hpxmap = hspmap.generate_healpix_map()

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxmap(hpxmap, nest=True)
    sp.draw_inset_colorbar()
    # These should match the healsparse maps, so we can use the same comparison.
    fname = 'healsparse_one.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    sp.draw_hpxmap(hpxmap, nest=True, lon_range=[4.9, 5.3], lat_range=[4.9, 5.3],
                   cmap=plt.colormaps['rainbow'])
    sp.draw_colorbar()
    fname = 'healsparse_two.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.DESSkyproj(ax=ax)
    sp.draw_hpxmap(hpxmap, nest=True)
    sp.draw_inset_colorbar()
    fname = 'healsparse_four.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_healpix_rasterized(tmp_path):
    """Test plotting a healpix map with and without rasterization."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()
    hpxmap = hspmap.generate_healpix_map()

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    # We use a small xsize here because otherwise the non-rasterized test pdf
    # (below) is 22Mb and the test takes a long time to run.
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxmap(hpxmap, nest=True, rasterized=True, xsize=200)
    sp.draw_inset_colorbar()
    fname = 'healpix_rasterized_on.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_on = os.path.getsize(tmp_path / fname)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxmap(hpxmap, nest=True, rasterized=False, xsize=200)
    sp.draw_inset_colorbar()
    fname = 'healpix_rasterized_off.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_off = os.path.getsize(tmp_path / fname)

    # The rasterized map will be smaller than the non-rasterized map.
    assert size_rasterized_on < size_rasterized_off

    # Test with a redraw event.
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxmap(hpxmap, nest=True, rasterized=True, xsize=200)
    sp.set_extent([10.5, -0.5, -0.5, 10.5])
    sp.draw_inset_colorbar()
    fname = 'healsparse_rasterized_on2.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_on2 = os.path.getsize(tmp_path / fname)

    # The rasterized map will be smaller than the non-rasterized map.
    assert size_rasterized_on2 < size_rasterized_off


def test_healsparse_widemask(tmp_path):
    """Test plotting a healsparse wide mask."""
    plt.rcParams.update(plt.rcParamsDefault)

    # Start with a 1-byte width map
    hspmap = hsp.HealSparseMap.make_empty(32, 4096, hsp.WIDE_MASK, wide_mask_maxbits=7)
    poly = hsp.geom.Polygon(ra=[0.0, 10.0, 10.0, 0.0], dec=[0.0, 0.0, 10.0, 10.0], value=1.0)
    pixels = poly.get_pixels(nside=hspmap.nside_sparse)
    hspmap.set_bits_pix(pixels, [0])
    poly2 = hsp.geom.Polygon(ra=[5, 5.2, 5.2, 5.0], dec=[5, 5.0, 5.2, 5.2], value=3.0)
    pixels2 = poly2.get_pixels(nside=hspmap.nside_sparse)
    hspmap.set_bits_pix(pixels2, [4])

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname = 'healsparse_wide_one.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    # And do the same with a 2-byte wide map
    hspmap = hsp.HealSparseMap.make_empty(32, 4096, hsp.WIDE_MASK, wide_mask_maxbits=15)
    poly = hsp.geom.Polygon(ra=[0.0, 10.0, 10.0, 0.0], dec=[0.0, 0.0, 10.0, 10.0], value=1.0)
    pixels = poly.get_pixels(nside=hspmap.nside_sparse)
    hspmap.set_bits_pix(pixels, [0])
    poly2 = hsp.geom.Polygon(ra=[5, 5.2, 5.2, 5.0], dec=[5, 5.0, 5.2, 5.2], value=3.0)
    pixels2 = poly2.get_pixels(nside=hspmap.nside_sparse)
    hspmap.set_bits_pix(pixels2, [10])

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname2 = 'healsparse_wide_one_alt.png'
    fig.savefig(tmp_path / fname2)
    # Compare to the previoues one.
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname2, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_healsparse_rec_array(tmp_path):
    """Test plotting a healsparse rec_array map."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap_rec_array()

    # Plot one component of map
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap["A"])
    sp.draw_inset_colorbar()
    fname = 'healsparse_rec_array.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    # Plot valid pixels of map
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap, valid_mask=True)
    sp.draw_inset_colorbar()
    fname = 'healsparse_valid_pixels.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_healsparse_rec_array_failover(tmp_path):
    """Test plotting the healsparse map without specifying a component
    Should get a UserWarning and a plot matching valid_mask=True.
    """
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap_rec_array()

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    with pytest.warns(UserWarning):
        im, lon_raster, lat_raster, values_raster = sp.draw_hspmap(hspmap)
    sp.draw_inset_colorbar()
    fname = 'healsparse_valid_pixels.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_hpxpix(tmp_path):
    """Test plotting healpix pixels."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()
    pixels = hspmap.valid_pixels
    values = hspmap[pixels]

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxpix(hspmap.nside_sparse,
                                                               pixels,
                                                               values,
                                                               nest=True)
    sp.draw_inset_colorbar()
    fname = 'healsparse_one.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    sp.draw_hpxpix(hspmap.nside_sparse, pixels, values, nest=True,
                   lon_range=[4.9, 5.3], lat_range=[4.9, 5.3],
                   cmap=plt.colormaps['rainbow'])
    sp.draw_colorbar()
    fname = 'healsparse_two.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.DESSkyproj(ax=ax)
    sp.draw_hpxpix(hspmap.nside_sparse, pixels, values, nest=True)
    sp.draw_inset_colorbar()
    fname = 'healsparse_four.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)


def test_hpxpix_rasterized(tmp_path):
    """Test plotting healpix pixels with and without rasterization."""
    plt.rcParams.update(plt.rcParamsDefault)

    hspmap = _get_hspmap()
    pixels = hspmap.valid_pixels
    values = hspmap[pixels]

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    # We use a small xsize here because otherwise the non-rasterized test pdf
    # (below) is 22Mb and the test takes a long time to run.
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxpix(hspmap.nside_sparse,
                                                               pixels,
                                                               values,
                                                               nest=True,
                                                               rasterized=True,
                                                               xsize=200)
    sp.draw_inset_colorbar()
    fname = 'hpxpix_rasterized_on.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_on = os.path.getsize(tmp_path / fname)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    # We use a small xsize here because otherwise the non-rasterized test pdf
    # (below) is 22Mb and the test takes a long time to run.
    im, lon_raster, lat_raster, values_raster = sp.draw_hpxpix(hspmap.nside_sparse,
                                                               pixels,
                                                               values,
                                                               nest=True,
                                                               rasterized=False,
                                                               xsize=200)
    sp.draw_inset_colorbar()
    fname = 'hpxpix_rasterized_off.pdf'
    fig.savefig(tmp_path / fname)

    size_rasterized_off = os.path.getsize(tmp_path / fname)

    # The rasterized map will be smaller than the non-rasterized map.
    assert size_rasterized_on < size_rasterized_off



import os

import numpy as np
import hpgeom as hpg

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


def test_healpix_binning(tmp_path):
    """Test healpix binning functionality."""
    plt.rcParams.update(plt.rcParamsDefault)

    np.random.seed(1234)

    ra = np.random.uniform(low=30.0, high=40.0, size=10000)
    dec = np.random.uniform(low=-55.0, high=-45.0, size=10000)
    C = np.random.uniform(low=0.0, high=10.0, size=10000)

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    hpxmap, im, lon_raster, lat_raster, values_raster = sp.draw_hpxbin(ra, dec)

    # Spot-check a pixel
    pix = hpg.angle_to_pixel(hpg.npixel_to_nside(hpxmap.size), ra, dec, nest=False)
    test, = np.where(pix == 671795)
    assert hpxmap[671795] == test.size

    fname = 'hpxbin.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)

    # Redo with averaging over values
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.McBrydeSkyproj(ax=ax)
    hpxmap, im, lon_raster, lat_raster, values_raster = sp.draw_hpxbin(ra, dec, C=C)

    # Spot-check the pixel
    np.testing.assert_approx_equal(hpxmap[671795], np.mean(C[test]))

    # And do with a survey
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.DESSkyproj(ax=ax)
    hpxmap, im, lon_raster, lat_raster, values_raster = sp.draw_hpxbin(ra, dec)
    fname = 'hpxbin_two.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 40.0)
    if err:
        raise ImageComparisonFailure(err)



import matplotlib.transforms
from matplotlib.path import Path
import numpy as np
from pyproj import Geod

from .skycrs import PlateCarreeCRS
from .utils import wrap_values

__all__ = ["SkyTransform"]


class SkyTransform(matplotlib.transforms.Transform):
    """Class for transforming to and from sky coordinates.

    This class implements the `matplotlib.transforms.Transform` API
    to convert to and from the rectilinear Plate caree (lon/lat)
    coordinate space to a specified coordinate system.  It does not
    do general transformations between two arbitrary coordinate systems.

    Parameters
    ----------
    proj : `skyproj.SkyProjection`
        Sky projection to transform from/to.
    inverse : `bool`, optional
        If ``False``, the transform will convert from lon/lat to the
        coordinate system ``proj``.  If ``True``, the transform will
        convert from ``proj`` to lon/lat.
    """
    input_dims = 2
    output_dims = 2
    is_separable = False
    has_inverse = True

    def __init__(self, proj, inverse=False):
        self._inverse = inverse
        if not inverse:
            self.source_proj = PlateCarreeCRS()
            self.target_proj = proj
        else:
            self.source_proj = proj
            self.target_proj = PlateCarreeCRS()

        # Number of geodesic sub-samples for paths.
        self._nsamp = 10
        self._nsamp_resolve = 50
        self._geod = Geod(a=self.source_proj.radius)

        self._lon_0 = self.target_proj.proj4_params['lon_0']
        self._wrap = (self._lon_0 + 180.) % 360.

        super().__init__()

    def inverted(self):
        # docstring inherited
        if not self._inverse:
            # Return the inverse
            return SkyTransform(self.target_proj, inverse=True)
        else:
            # Re-invert it
            return SkyTransform(self.source_proj, inverse=False)

    def transform_non_affine(self, xy):
        # docstring inherited
        res = self.target_proj.transform_points(self.source_proj,
                                                xy[:, 0], xy[:, 1])

        return res

    def transform_path_non_affine(self, path):
        # docstring inherited
        if self._inverse:
            # Just send this upstream if we're not computing geodesics.
            return super().transform_path_non_affine(path)

        if self.target_proj.name == 'obmoll':
            return self._transform_path_non_affine_oblique(path)

        lonlats = []
        codes = []

        last_vertex = None
        is_polygon = False
        # Make sure we don't simplify the path segments, which gets coordinate
        # systems all messed up.
        for vertex, code in path.iter_segments(simplify=False):
            if last_vertex is None or code == Path.MOVETO:
                lonlats.extend([(vertex[0], vertex[1])])
                codes.append(Path.MOVETO)
                last_vertex = vertex
            elif code in (Path.LINETO, Path.CLOSEPOLY, None):
                # Connect the last vertex
                lonlats_step = self._geod.npts(last_vertex[0], last_vertex[1],
                                               vertex[0], vertex[1], self._nsamp + 1,
                                               initial_idx=1, terminus_idx=0)
                lonlats.extend(lonlats_step)
                if code == Path.CLOSEPOLY:
                    is_polygon = True
                codes.extend([Path.LINETO]*len(lonlats_step))
                last_vertex = vertex
            else:
                raise ValueError("Unsupported code type %d" % (code))

        lonlats = np.array(lonlats)
        codes = np.array(codes)

        # Normalize range
        lonlats[:, 0] = wrap_values(lonlats[:, 0], wrap=self._wrap)

        # Cut into segments that wrap around
        cuts = self._compute_cuts(lonlats[:, 0])

        if cuts.size > 0:
            # Modify path to hit the edge and jump to the other side.
            lonlats, codes = self._insert_jumps(lonlats, codes, cuts)

            if is_polygon:
                # If this is a polygon we need to add extra edges.
                lonlats, codes = self._complete_cut_polygons(lonlats, codes)

        vertices_xform = self.target_proj.transform_points(self.source_proj,
                                                           lonlats[:, 0], lonlats[:, 1])

        new_path = Path(vertices_xform, codes)

        return new_path

    def _compute_cuts(self, lons):
        """Compute cut locations from a list of longitudes.

        Parameters
        ----------
        lons : `np.ndarray`
            Array of longitude values.

        Returns
        -------
        cuts : `np.ndarray`
            Array of jump/cut locations.
        """
        cuts, = np.where(np.abs(lons[: -1] - lons[1:]) > 180.0)
        return cuts

    def _insert_jumps(self, lonlats, codes, cuts):
        """Insert jump points at edges.

        Parameters
        ----------
        lonlats : `np.ndarray`
            [N, 2] array of longitude/latitudes
        codes : `np.ndarray`
            Array of path codes.
        cuts : `np.ndarray`
            Array of cut locations

        Returns
        -------
        lonlats : `np.ndarray`
            New [M, 2] array of longitude/latitudes
        codes : `np.ndarray`
            New array of path codes.
        """
        lonlats_insert = []
        locs_insert = []
        codes_insert = []
        for c in cuts:
            jump_vals = [self._wrap - 1e-10, self._wrap - 360. + 1e-10]
            if lonlats[c, 0] < self._lon_0:
                # Switch the order if it's at the low end.
                jump_vals = jump_vals[::-1]

            lonlats_insert.extend([(jump_vals[0], lonlats[c, 1])])
            lonlats_insert.extend([(jump_vals[1], lonlats[c + 1, 1])])
            locs_insert.extend([c + 1]*2)
            codes_insert.extend([Path.LINETO, Path.MOVETO])

        lonlats = np.insert(lonlats, locs_insert, lonlats_insert, axis=0)
        codes = np.insert(codes, locs_insert, codes_insert)

        return lonlats, codes

    def _complete_cut_polygons(self, lonlats, codes):
        """Complete cut polygon shapes.

        Parameters
        ----------
        lonlats : `np.ndarray`
            [N, 2] array of lonlats.
        codes : `np.ndarray`
            Array of path codes.

        Returns
        -------
        lonlats : `np.ndarray`
            New [M, 2] array of longitude/latitudes
        codes : `np.ndarray`
            New array of path codes.
        """
        # First, we need to roll to the first cut at the edge.
        cuts = self._compute_cuts(lonlats[:, 0])
        lonlats = np.roll(lonlats, -(cuts[0] + 1), axis=0)
        codes = np.roll(codes, -(cuts[0] + 1))

        # Recompute cuts after roll
        cuts = self._compute_cuts(lonlats[:, 0])

        # We need to handle inserting and appending separately, apparently.
        poly_vertex_start = lonlats[0, :]

        if cuts.size > 0:
            lonlats_insert = []
            locs_insert = []
            codes_insert = []
            for c in cuts:
                lonlats_step = self._geod.npts(lonlats[c - 1, 0], lonlats[c - 1, 1],
                                               poly_vertex_start[0], poly_vertex_start[1],
                                               self._nsamp + 1,
                                               initial_idx=1, terminus_idx=0)
                lonlats_insert.extend(lonlats_step)
                locs_insert.extend([c + 1]*len(lonlats_step))
                codes_insert.extend([Path.LINETO]*len(lonlats_step))
                poly_vertex_start = lonlats[c + 1, :]

            lonlats_insert = np.array(lonlats_insert)
            lonlats_insert[:, 0] = wrap_values(lonlats_insert[:, 0], wrap=self._wrap)

            lonlats = np.insert(lonlats, locs_insert, lonlats_insert, axis=0)
            codes = np.insert(codes, locs_insert, codes_insert)

        # And the final connection
        lonlats_step = self._geod.npts(lonlats[-1, 0], lonlats[-1, 1],
                                       poly_vertex_start[0], poly_vertex_start[1],
                                       self._nsamp + 1,
                                       initial_idx=1, terminus_idx=0)
        lonlats_append = np.array(lonlats_step)
        lonlats_append[:, 0] = wrap_values(lonlats_append[:, 0], wrap=self._wrap)
        codes_append = [Path.LINETO]*len(lonlats_append)

        lonlats = np.append(lonlats, lonlats_append, axis=0)
        codes = np.append(codes, codes_append)

        return lonlats, codes

    def _transform_path_non_affine_oblique(self, path):
        """Transform a non-affine path for an oblique transformation.

        Parameters
        ----------
        path : `matplotlib.path.Path`
            Path to transform.

        Returns
        -------
        transformed_path : `matplotlib.path.Path`
            Path transformed to new coordinate system.
        """
        # In the future, this routine may be merged with the one above.
        # However, there are significant challenges in tracking the edge
        # of the warp for oblique transforms which is necessary to create
        # wrapped filled polygons.
        vertices_xform = []
        codes = []

        last_vertex = None
        for vertex, code in path.iter_segments(simplify=False):
            if code == Path.MOVETO:
                vertex_xform = self.target_proj.transform_points(self.source_proj,
                                                                 vertex[0],
                                                                 vertex[1])
                vertices_xform.extend([(vertex_xform[0][0], vertex_xform[0][1])])
                codes.append(Path.MOVETO)
                last_vertex = vertex
            elif code in (Path.LINETO, None):
                # Connect the last vertex
                lonlats_step = np.array(self._geod.npts(last_vertex[0], last_vertex[1],
                                                        vertex[0], vertex[1], self._nsamp + 1,
                                                        initial_idx=0, terminus_idx=0))

                lonlats_step_xform = self.target_proj.transform_points(self.source_proj,
                                                                       lonlats_step[:, 0],
                                                                       lonlats_step[:, 1])

                dxdy = lonlats_step_xform[1:, :] - lonlats_step_xform[0: -1, :]
                split, = (np.hypot(dxdy[:, 0], dxdy[:, 1]) > 0.5*self.target_proj.radius).nonzero()

                if split.size > 0:
                    if split.size > 1:
                        raise NotImplementedError("Cannot plot a line with more than one cut.")

                    index = split[0] + 1

                    lonlats_temp = np.array(self._geod.npts(lonlats_step[index - 1, 0],
                                                            lonlats_step[index - 1, 1],
                                                            lonlats_step[index, 0],
                                                            lonlats_step[index, 1],
                                                            self._nsamp_resolve + 1,
                                                            initial_idx=1, terminus_idx=0))
                    lonlats_temp_xform = self.target_proj.transform_points(self.source_proj,
                                                                           lonlats_temp[:, 0],
                                                                           lonlats_temp[:, 1])
                    dxdy_temp = lonlats_temp_xform[1:, :] - lonlats_temp_xform[0: -1, :]
                    split_temp, = (np.hypot(dxdy_temp[:, 0], dxdy_temp[:, 1])
                                   > 0.5*self.target_proj.radius).nonzero()

                    index_temp = split_temp[0] + 1

                    # Pre-step
                    vertices_xform.extend(lonlats_step_xform[: index, :].tolist())
                    codes.extend([Path.LINETO]*index)
                    vertices_xform.extend(lonlats_temp_xform[: index_temp, :].tolist())
                    codes.extend([Path.LINETO]*index_temp)
                    vertices_xform.extend(lonlats_temp_xform[index_temp:, :].tolist())
                    codes.extend([Path.MOVETO])
                    codes.extend([Path.LINETO]*(self._nsamp_resolve - index_temp))
                    vertices_xform.extend(lonlats_step_xform[index:, :].tolist())
                    codes.extend([Path.LINETO]*(self._nsamp + 1 - index))
                else:
                    vertices_xform.extend(lonlats_step_xform.tolist())
                    codes.extend([Path.LINETO]*(self._nsamp + 1))

                last_vertex = vertex

        new_path = Path(np.array(vertices_xform), codes)

        return new_path



try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("skyproj")
except PackageNotFoundError:
    # package is not installed
    pass

from .skyproj import *
from .skycrs import *
from .transforms import *
from .skyaxes import *
from .survey import *
